#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2019-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import contextlib
import json
import os
import pathlib
import tempfile
import unittest
from unittest import mock


from edgedb import con_utils
from edgedb import errors


class TestConUtils(unittest.TestCase):

    TESTS = [
        {
            'user': 'user',
            'host': 'localhost',
            'result': (
                [("localhost", 5656)],
                {
                    'user': 'user',
                    'database': 'edgedb',
                },
                {'wait_until_available': 30},
            )
        },

        {
            'env': {
                'EDGEDB_USER': 'user',
                'EDGEDB_DATABASE': 'testdb',
                'EDGEDB_PASSWORD': 'passw',
                'EDGEDB_HOST': 'host',
                'EDGEDB_PORT': '123'
            },
            'result': (
                [('host', 123)],
                {
                    'user': 'user',
                    'password': 'passw',
                    'database': 'testdb'
                },
                {'wait_until_available': 30},
            )
        },

        {
            'env': {
                'EDGEDB_USER': 'user',
                'EDGEDB_DATABASE': 'testdb',
                'EDGEDB_PASSWORD': 'passw',
                'EDGEDB_HOST': 'host',
                'EDGEDB_PORT': '123'
            },

            'host': 'host2',
            'port': '456',
            'user': 'user2',
            'password': 'passw2',
            'database': 'db2',

            'result': (
                [('host2', 456)],
                {
                    'user': 'user2',
                    'password': 'passw2',
                    'database': 'db2'
                },
                {'wait_until_available': 30},
            )
        },

        {
            'env': {
                'EDGEDB_USER': 'user',
                'EDGEDB_DATABASE': 'testdb',
                'EDGEDB_PASSWORD': 'passw',
                'EDGEDB_HOST': 'host',
                'EDGEDB_PORT': '123',
                'PGSSLMODE': 'prefer'
            },

            'dsn': 'edgedb://user3:123123@localhost/abcdef',

            'host': 'host2',
            'port': '456',
            'user': 'user2',
            'password': 'passw2',
            'database': 'db2',
            'server_settings': {'ssl': 'False'},

            'result': (
                [('host2', 456)],
                {
                    'user': 'user2',
                    'password': 'passw2',
                    'database': 'db2',
                    'server_settings': {'ssl': 'False'},
                },
                {'wait_until_available': 30},
            )
        },

        {
            'env': {
                'EDGEDB_USER': 'user',
                'EDGEDB_DATABASE': 'testdb',
                'EDGEDB_PASSWORD': 'passw',
                'EDGEDB_HOST': 'host',
                'EDGEDB_PORT': '123',
            },

            'dsn': 'edgedb://user3:123123@localhost:5555/abcdef',
            'command_timeout': 10,

            'result': (
                [('localhost', 5555)],
                {
                    'user': 'user3',
                    'password': '123123',
                    'database': 'abcdef',
                }, {
                    'command_timeout': 10,
                    'wait_until_available': 30,
                })
        },

        {
            'dsn': 'edgedb://user3:123123@localhost:5555/abcdef',
            'result': (
                [('localhost', 5555)],
                {
                    'user': 'user3',
                    'password': '123123',
                    'database': 'abcdef'
                },
                {'wait_until_available': 30},
            )
        },

        {
            'dsn': 'edgedb://user@host1,host2/db',
            'result': (
                [('host1', 5656), ('host2', 5656)],
                {
                    'database': 'db',
                    'user': 'user',
                },
                {'wait_until_available': 30},
            )
        },

        {
            'dsn': 'edgedb://user@host1:1111,host2:2222/db',
            'result': (
                [('host1', 1111), ('host2', 2222)],
                {
                    'database': 'db',
                    'user': 'user',
                },
                {'wait_until_available': 30},
            )
        },

        {
            'env': {
                'EDGEDB_HOST': 'host1:1111,host2:2222',
                'EDGEDB_USER': 'foo',
            },
            'dsn': 'edgedb:///db',
            'result': (
                [('host1', 1111), ('host2', 2222)],
                {
                    'database': 'db',
                    'user': 'foo',
                },
                {'wait_until_available': 30},
            )
        },

        {
            'env': {
                'EDGEDB_USER': 'foo',
            },
            'dsn': 'edgedb:///db?host=host1:1111,host2:2222',
            'result': (
                [('host1', 1111), ('host2', 2222)],
                {
                    'database': 'db',
                    'user': 'foo',
                },
                {'wait_until_available': 30},
            )
        },

        {
            'env': {
                'EDGEDB_USER': 'foo',
            },
            'dsn': 'edgedb:///db',
            'host': ['host1', 'host2'],
            'result': (
                [('host1', 5656), ('host2', 5656)],
                {
                    'database': 'db',
                    'user': 'foo',
                },
                {'wait_until_available': 30},
            )
        },

        {
            'dsn': 'edgedb://user3:123123@localhost:5555/'
                   'abcdef?param=sss&param=123&host=testhost&user=testuser'
                   '&port=2222&database=testdb',
            'host': '127.0.0.1',
            'port': '888',
            'user': 'me',
            'password': 'ask',
            'database': 'db',
            'result': (
                [('127.0.0.1', 888)],
                {
                    'server_settings': {'param': '123'},
                    'user': 'me',
                    'password': 'ask',
                    'database': 'db',
                },
                {'wait_until_available': 30},
            )
        },

        {
            'dsn': 'edgedb://user3:123123@localhost:5555/'
                   'abcdef?param=sss&param=123&host=testhost&user=testuser'
                   '&port=2222&database=testdb',
            'host': '127.0.0.1',
            'port': '888',
            'user': 'me',
            'password': 'ask',
            'database': 'db',
            'server_settings': {'aa': 'bb'},
            'result': (
                [('127.0.0.1', 888)],
                {
                    'server_settings': {'aa': 'bb', 'param': '123'},
                    'user': 'me',
                    'password': 'ask',
                    'database': 'db',
                },
                {'wait_until_available': 30},
            )
        },

        {
            'dsn': 'edgedb:///dbname?host=/unix_sock/test&user=spam',
            'result': (
                [os.path.join('/unix_sock/test', '.s.EDGEDB.5656')],
                {
                    'user': 'spam',
                    'database': 'dbname'
                },
                {'wait_until_available': 30},
            )
        },

        {
            'dsn': 'pq:///dbname?host=/unix_sock/test&user=spam',
            'error': (
                ValueError,
                "dsn "
                "'pq:///dbname\\?host=/unix_sock/test&user=spam' "
                "is neither a edgedb:// URI nor valid instance name"
            )
        },

        {
            'dsn': 'edgedb://host1,host2,host3/db',
            'port': [111, 222],
            'error': (
                errors.InterfaceError,
                'could not match 2 port numbers to 3 hosts'
            )
        },

        {
            'dsn': 'edgedb://user@?port=56226&host=%2Ftmp',
            'result': (
                [os.path.join('/tmp', '.s.EDGEDB.56226')],
                {
                    'user': 'user',
                    'database': 'edgedb',
                },
                {'wait_until_available': 30},
            )
        },

        {
            'dsn': 'edgedb://user@?host=%2Ftmp',
            'admin': True,
            'result': (
                [os.path.join('/tmp', '.s.EDGEDB.admin.5656')],
                {
                    'user': 'user',
                    'database': 'edgedb',
                },
                {'wait_until_available': 30},
            )
        },

        {
            'dsn': 'edgedbadmin://user@?host=%2Ftmp',
            'result': (
                [os.path.join('/tmp', '.s.EDGEDB.admin.5656')],
                {
                    'user': 'user',
                    'database': 'edgedb',
                },
                {'wait_until_available': 30},
            )
        },
    ]

    @contextlib.contextmanager
    def environ(self, **kwargs):
        old_vals = {}
        for key in kwargs:
            if key in os.environ:
                old_vals[key] = os.environ[key]

        for key, val in kwargs.items():
            if val is None:
                if key in os.environ:
                    del os.environ[key]
            else:
                os.environ[key] = val

        try:
            yield
        finally:
            for key in kwargs:
                if key in os.environ:
                    del os.environ[key]
            for key, val in old_vals.items():
                os.environ[key] = val

    def run_testcase(self, testcase):
        env = testcase.get('env', {})
        test_env = {'EDGEDB_HOST': None, 'EDGEDB_PORT': None,
                    'EDGEDB_USER': None, 'EDGEDB_PASSWORD': None,
                    'EDGEDB_DATABASE': None, 'PGSSLMODE': None}
        test_env.update(env)

        dsn = testcase.get('dsn')
        user = testcase.get('user')
        port = testcase.get('port')
        host = testcase.get('host')
        password = testcase.get('password')
        database = testcase.get('database')
        tls_ca_file = testcase.get('tls_ca_file')
        tls_verify_hostname = testcase.get('tls_verify_hostname')
        admin = testcase.get('admin')
        timeout = testcase.get('timeout')
        command_timeout = testcase.get('command_timeout')
        server_settings = testcase.get('server_settings')

        expected = testcase.get('result')
        expected_error = testcase.get('error')
        if expected is None and expected_error is None:
            raise RuntimeError(
                'invalid test case: either "result" or "error" key '
                'has to be specified')
        if expected is not None and expected_error is not None:
            raise RuntimeError(
                'invalid test case: either "result" or "error" key '
                'has to be specified, got both')

        result = None
        with contextlib.ExitStack() as es:
            es.enter_context(self.subTest(dsn=dsn, env=env))
            es.enter_context(self.environ(**test_env))

            if expected_error:
                es.enter_context(self.assertRaisesRegex(*expected_error))

            addrs, params, config = con_utils.parse_connect_arguments(
                dsn=dsn, host=host, port=port, user=user, password=password,
                database=database, admin=admin,
                tls_ca_file=tls_ca_file,
                tls_verify_hostname=tls_verify_hostname,
                timeout=timeout, command_timeout=command_timeout,
                server_settings=server_settings,
                wait_until_available=30)

            params = {k: v for k, v in params._asdict().items()
                      if v is not None}
            config = {k: v for k, v in config._asdict().items()
                      if v is not None}
            params.pop('ssl_ctx', None)

            result = (addrs, params, config)

        if expected is not None:
            for k, v in expected[1].items():
                # If `expected` contains a type, allow that to "match" any
                # instance of that type that `result` may contain.
                if isinstance(v, type) and isinstance(result[1].get(k), v):
                    result[1][k] = v
            self.assertEqual(expected, result, 'Testcase: {}'.format(testcase))

    def test_test_connect_params_environ(self):
        self.assertNotIn('AAAAAAAAAA123', os.environ)
        self.assertNotIn('AAAAAAAAAA456', os.environ)
        self.assertNotIn('AAAAAAAAAA789', os.environ)

        try:

            os.environ['AAAAAAAAAA456'] = '123'
            os.environ['AAAAAAAAAA789'] = '123'

            with self.environ(AAAAAAAAAA123='1',
                              AAAAAAAAAA456='2',
                              AAAAAAAAAA789=None):

                self.assertEqual(os.environ['AAAAAAAAAA123'], '1')
                self.assertEqual(os.environ['AAAAAAAAAA456'], '2')
                self.assertNotIn('AAAAAAAAAA789', os.environ)

            self.assertNotIn('AAAAAAAAAA123', os.environ)
            self.assertEqual(os.environ['AAAAAAAAAA456'], '123')
            self.assertEqual(os.environ['AAAAAAAAAA789'], '123')

        finally:
            for key in {'AAAAAAAAAA123', 'AAAAAAAAAA456', 'AAAAAAAAAA789'}:
                if key in os.environ:
                    del os.environ[key]

    def test_test_connect_params_run_testcase(self):
        with self.environ(EDGEDB_PORT='777'):
            self.run_testcase({
                'env': {
                    'EDGEDB_USER': '__test__'
                },
                'host': 'abc',
                'result': (
                    [('abc', 5656)],
                    {'user': '__test__', 'database': 'edgedb'},
                    {'wait_until_available': 30},
                )
            })

    def test_connect_params(self):
        for testcase in self.TESTS:
            self.run_testcase(testcase)

    @mock.patch("edgedb.platform.config_dir",
                lambda: pathlib.Path("/home/user/.config/edgedb"))
    @mock.patch("edgedb.platform.IS_WINDOWS", False)
    @mock.patch("pathlib.Path.exists", lambda p: True)
    @mock.patch("os.path.realpath", lambda p: p)
    def test_stash_path(self):
        self.assertEqual(
            con_utils._stash_path("/home/user/work/project1"),
            pathlib.Path("/home/user/.config/edgedb/projects/project1-"
                         "cf1c841351bf7f147d70dcb6203441cf77a05249"),
        )

    def test_project_config(self):
        with tempfile.TemporaryDirectory() as tmp:
            tmp = os.path.realpath(tmp)
            base = pathlib.Path(tmp)
            home = base / 'home'
            project = base / 'project'
            projects = home / '.edgedb' / 'projects'
            creds = home / '.edgedb' / 'credentials'
            os.makedirs(projects)
            os.makedirs(creds)
            os.makedirs(project)
            with open(project / 'edgedb.toml', 'wt') as f:
                f.write('')  # app don't read toml file
            with open(creds / 'inst1.json', 'wt') as f:
                f.write(json.dumps({
                    "host": "inst1.example.org",
                    "port": 12323,
                    "user": "inst1_user",
                    "password": "passw1",
                    "database": "inst1_db",
                }))

            with mock.patch('edgedb.platform.config_dir',
                            lambda: home / '.edgedb'), \
                    mock.patch('os.getcwd', lambda: str(project)):
                stash_path = con_utils._stash_path(project)
                instance_file = stash_path / 'instance-name'
                os.makedirs(stash_path)
                with open(instance_file, 'wt') as f:
                    f.write('inst1')

                addrs, params, _config = con_utils.parse_connect_arguments(
                    dsn=None, host=None, port=None, user=None, password=None,
                    database=None, admin=None,
                    tls_ca_file=None, tls_verify_hostname=None,
                    timeout=10, command_timeout=None,
                    server_settings=None,
                    wait_until_available=30)

        self.assertEqual(addrs, [('inst1.example.org', 12323)])
        self.assertEqual(params.user, 'inst1_user')
        self.assertEqual(params.password, 'passw1')
        self.assertEqual(params.database, 'inst1_db')



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2019-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


cdef class EnumDescriptor:
    cdef:
        object tid
        dict index
        tuple labels

    cdef get_index(self, EnumValue v)


cdef class EnumValue:
    cdef:
        EnumDescriptor desc
        str label

    cdef get_index(self)



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2016-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module provides support for legacy `connection.transaction()` method.

This API is deprecated and will eventually be removed.
Use `retrying_transaction()` or `raw_transaction()` instead.
"""

import enum

from . import errors


__all__ = ('Transaction', 'AsyncIOTransaction')


class TransactionState(enum.Enum):
    NEW = 0
    STARTED = 1
    COMMITTED = 2
    ROLLEDBACK = 3
    FAILED = 4


ISOLATION_LEVELS = {'serializable', 'repeatable_read'}


class BaseTransaction:

    __slots__ = ('_connection', '_isolation', '_readonly', '_deferrable',
                 '_state', '_nested', '_id', '_managed')

    def __init__(self, connection, isolation: str,
                 readonly: bool, deferrable: bool):
        if isolation is not None and isolation not in ISOLATION_LEVELS:
            raise ValueError(
                'isolation is expected to be either of {}, '
                'got {!r}'.format(ISOLATION_LEVELS, isolation))

        self._connection = connection
        self._isolation = isolation
        self._readonly = readonly
        self._deferrable = deferrable
        self._state = TransactionState.NEW
        self._nested = False
        self._id = None
        self._managed = False

    def is_active(self) -> bool:
        return self._state is TransactionState.STARTED

    def __check_state_base(self, opname):
        if self._state is TransactionState.COMMITTED:
            raise errors.InterfaceError(
                'cannot {}; the transaction is already committed'.format(
                    opname))
        if self._state is TransactionState.ROLLEDBACK:
            raise errors.InterfaceError(
                'cannot {}; the transaction is already rolled back'.format(
                    opname))
        if self._state is TransactionState.FAILED:
            raise errors.InterfaceError(
                'cannot {}; the transaction is in error state'.format(
                    opname))

    def __check_state(self, opname):
        if self._state is not TransactionState.STARTED:
            if self._state is TransactionState.NEW:
                raise errors.InterfaceError(
                    'cannot {}; the transaction is not yet started'.format(
                        opname))
            self.__check_state_base(opname)

    def _make_start_query(self):
        self.__check_state_base('start')
        if self._state is TransactionState.STARTED:
            raise errors.InterfaceError(
                'cannot start; the transaction is already started')

        con = self._connection_inner

        if con._top_xact is None:
            con._top_xact = self
        else:
            # Nested transaction block
            top_xact = con._top_xact
            if self._isolation is None:
                self._isolation = top_xact._isolation
            if self._readonly is None:
                self._readonly = top_xact._readonly
            if self._deferrable is None:
                self._deferrable = top_xact._deferrable

            if self._isolation != top_xact._isolation:
                raise errors.InterfaceError(
                    'nested transaction has a different isolation level: '
                    'current {!r} != outer {!r}'.format(
                        self._isolation, top_xact._isolation))

            if self._readonly != top_xact._readonly:
                raise errors.InterfaceError(
                    'nested transaction has a different read-write spec: '
                    'current {!r} != outer {!r}'.format(
                        self._readonly, top_xact._readonly))

            if self._deferrable != top_xact._deferrable:
                raise errors.InterfaceError(
                    'nested transaction has a different deferrable spec: '
                    'current {!r} != outer {!r}'.format(
                        self._deferrable, top_xact._deferrable))

            self._nested = True

        if self._nested:
            self._id = con._get_unique_id('savepoint')
            query = f'DECLARE SAVEPOINT {self._id};'
        else:
            query = 'START TRANSACTION'

            if self._isolation == 'repeatable_read':
                query = 'START TRANSACTION ISOLATION REPEATABLE READ'
            elif self._isolation == 'serializable':
                query = 'START TRANSACTION ISOLATION SERIALIZABLE'

            if self._readonly:
                query += ' READ ONLY'
            elif self._readonly is not None:
                query += ' READ WRITE'
            if self._deferrable:
                query += ' DEFERRABLE'
            elif self._deferrable is not None:
                query += ' NOT DEFERRABLE'
            query += ';'

        return query

    def _make_commit_query(self):
        self.__check_state('commit')

        if self._connection_inner._top_xact is self:
            self._connection_inner._top_xact = None

        if self._nested:
            query = f'RELEASE SAVEPOINT {self._id};'
        else:
            query = 'COMMIT;'

        return query

    def _make_rollback_query(self):
        self.__check_state('rollback')

        if self._connection_inner._top_xact is self:
            self._connection_inner._top_xact = None

        if self._nested:
            query = f'ROLLBACK TO SAVEPOINT {self._id};'
        else:
            query = 'ROLLBACK;'

        return query

    def __repr__(self):
        attrs = []
        attrs.append('state:{}'.format(self._state.name.lower()))

        if self._isolation:
            attrs.append(self._isolation)
        if self._readonly:
            attrs.append('readonly')
        if self._deferrable:
            attrs.append('deferrable')

        if self.__class__.__module__.startswith('edgedb.'):
            mod = 'edgedb'
        else:
            mod = self.__class__.__module__

        return '<{}.{} {} {:#x}>'.format(
            mod, self.__class__.__name__, ' '.join(attrs), id(self))


class AsyncIOTransaction(BaseTransaction):

    async def __aenter__(self):
        if self._managed:
            raise errors.InterfaceError(
                'cannot enter context: already in an `async with` block')
        self._managed = True
        await self.start()

    async def __aexit__(self, extype, ex, tb):
        try:
            if extype is not None:
                await self.__rollback()
            else:
                await self.__commit()
        finally:
            self._managed = False

    async def start(self) -> None:
        """Enter the transaction or savepoint block."""
        await self._connection.ensure_connected()
        self._connection_inner = self._connection._inner
        self._connection_impl = self._connection_inner._impl

        query = self._make_start_query()
        try:
            await self._connection_impl.privileged_execute(query)
        except BaseException:
            self._state = TransactionState.FAILED
            raise
        else:
            self._state = TransactionState.STARTED

    async def __commit(self):
        query = self._make_commit_query()
        try:
            await self._connection_impl.privileged_execute(query)
        except BaseException:
            self._state = TransactionState.FAILED
            raise
        else:
            self._state = TransactionState.COMMITTED

    async def __rollback(self):
        query = self._make_rollback_query()
        try:
            await self._connection_impl.privileged_execute(query)
        except BaseException:
            self._state = TransactionState.FAILED
            raise
        else:
            self._state = TransactionState.ROLLEDBACK

    async def commit(self) -> None:
        """Exit the transaction or savepoint block and commit changes."""
        if self._managed:
            raise errors.InterfaceError(
                'cannot manually commit from within an `async with` block')
        await self.__commit()

    async def rollback(self) -> None:
        """Exit the transaction or savepoint block and rollback changes."""
        if self._managed:
            raise errors.InterfaceError(
                'cannot manually rollback from within an `async with` block')
        await self.__rollback()


class Transaction(BaseTransaction):

    def __enter__(self):
        if self._managed:
            raise errors.InterfaceError(
                'cannot enter context: already in a `with` block')
        self._managed = True
        self.start()

    def __exit__(self, extype, ex, tb):
        try:
            if extype is not None:
                self.__rollback()
            else:
                self.__commit()
        finally:
            self._managed = False

    def start(self) -> None:
        """Enter the transaction or savepoint block."""
        self._connection.ensure_connected()
        self._connection_inner = self._connection._inner
        self._connection_impl = self._connection_inner._impl
        query = self._make_start_query()
        try:
            self._connection_impl.privileged_execute(query)
        except BaseException:
            self._state = TransactionState.FAILED
            raise
        else:
            self._state = TransactionState.STARTED

    def __commit(self):
        query = self._make_commit_query()
        try:
            self._connection_impl.privileged_execute(query)
        except BaseException:
            self._state = TransactionState.FAILED
            raise
        else:
            self._state = TransactionState.COMMITTED

    def __rollback(self):
        query = self._make_rollback_query()
        try:
            self._connection_impl.privileged_execute(query)
        except BaseException:
            self._state = TransactionState.FAILED
            raise
        else:
            self._state = TransactionState.ROLLEDBACK

    def commit(self) -> None:
        """Exit the transaction or savepoint block and commit changes."""
        if self._managed:
            raise errors.InterfaceError(
                'cannot manually commit from within a `with` block')
        self.__commit()

    def rollback(self) -> None:
        """Exit the transaction or savepoint block and rollback changes."""
        if self._managed:
            raise errors.InterfaceError(
                'cannot manually rollback from within a `with` block')
        self.__rollback()



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2016-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import edgedb

from edgedb import _testbase as tb


class TestAsyncTxLegacy(tb.AsyncQueryTestCase):

    ISOLATED_METHODS = False

    SETUP = '''
        CREATE TYPE test::TransactionTest EXTENDING std::Object {
            CREATE PROPERTY name -> std::str;
        };
    '''

    TEARDOWN = '''
        DROP TYPE test::TransactionTest;
    '''

    async def test_async_transaction_regular_01(self):
        self.assertIsNone(self.con._inner._top_xact)
        tr = self.con.transaction()
        self.assertIsNone(self.con._inner._top_xact)

        with self.assertRaises(ZeroDivisionError):
            async with tr as with_tr:
                self.assertIs(self.con._inner._top_xact, tr)

                # We don't return the transaction object from __aenter__,
                # to make it harder for people to use '.rollback()' and
                # '.commit()' from within an 'async with' block.
                self.assertIsNone(with_tr)

                await self.con.execute('''
                    INSERT test::TransactionTest {
                        name := 'Test Transaction'
                    };
                ''')

                1 / 0

        self.assertIsNone(self.con._inner._top_xact)

        result = await self.con.query('''
            SELECT
                test::TransactionTest
            FILTER
                test::TransactionTest.name = 'Test Transaction';
        ''')

        self.assertEqual(result, [])

    async def test_async_transaction_nested_01(self):
        self.assertIsNone(self.con._inner._top_xact)
        tr = self.con.transaction()
        self.assertIsNone(self.con._inner._top_xact)

        with self.assertRaises(ZeroDivisionError):
            async with tr:
                self.assertIs(self.con._inner._top_xact, tr)

                async with self.con.transaction():
                    self.assertIs(self.con._inner._top_xact, tr)

                    await self.con.execute('''
                        INSERT test::TransactionTest {
                            name := 'TXTEST 1'
                        };
                    ''')

                self.assertIs(self.con._inner._top_xact, tr)

                with self.assertRaises(ZeroDivisionError):
                    in_tr = self.con.transaction()
                    async with in_tr:

                        self.assertIs(self.con._inner._top_xact, tr)

                        await self.con.execute('''
                            INSERT test::TransactionTest {
                                name := 'TXTEST 2'
                            };
                        ''')

                        1 / 0

                recs = await self.con.query('''
                    SELECT
                        test::TransactionTest {
                            name
                        }
                    FILTER
                        test::TransactionTest.name LIKE 'TXTEST%';
                ''')

                self.assertEqual(recs[0].name, 'TXTEST 1')
                self.assertIs(self.con._inner._top_xact, tr)

                1 / 0

        self.assertIs(self.con._inner._top_xact, None)

        recs = await self.con.query('''
            SELECT
                test::TransactionTest {
                    name
                }
            FILTER
                test::TransactionTest.name LIKE 'TXTEST%';
        ''')

        self.assertEqual(len(recs), 0)

    async def test_async_transaction_nested_02(self):
        async with self.con.transaction(isolation='repeatable_read'):
            async with self.con.transaction():  # no explicit isolation, OK
                pass

        with self.assertRaisesRegex(edgedb.InterfaceError,
                                    r'different isolation'):
            async with self.con.transaction(isolation='repeatable_read'):
                async with self.con.transaction(isolation='serializable'):
                    pass

        with self.assertRaisesRegex(edgedb.InterfaceError,
                                    r'different read-write'):
            async with self.con.transaction():
                async with self.con.transaction(readonly=True):
                    pass

        with self.assertRaisesRegex(edgedb.InterfaceError,
                                    r'different deferrable'):
            async with self.con.transaction(deferrable=True):
                async with self.con.transaction(deferrable=False):
                    pass

    async def test_async_transaction_interface_errors(self):
        self.assertIsNone(self.con._inner._top_xact)

        tr = self.con.transaction()
        with self.assertRaisesRegex(edgedb.InterfaceError,
                                    r'cannot start; .* already started'):
            async with tr:
                await tr.start()

        self.assertTrue(repr(tr).startswith(
            '<edgedb.AsyncIOTransaction state:rolledback'))

        self.assertIsNone(self.con._inner._top_xact)

        with self.assertRaisesRegex(edgedb.InterfaceError,
                                    r'cannot start; .* already rolled back'):
            async with tr:
                pass

        self.assertIsNone(self.con._inner._top_xact)

        tr = self.con.transaction()
        with self.assertRaisesRegex(edgedb.InterfaceError,
                                    r'cannot manually commit.*async with'):
            async with tr:
                await tr.commit()

        self.assertIsNone(self.con._inner._top_xact)

        tr = self.con.transaction()
        with self.assertRaisesRegex(edgedb.InterfaceError,
                                    r'cannot manually rollback.*async with'):
            async with tr:
                await tr.rollback()

        self.assertIsNone(self.con._inner._top_xact)

        tr = self.con.transaction()
        with self.assertRaisesRegex(edgedb.InterfaceError,
                                    r'cannot enter context:.*async with'):
            async with tr:
                async with tr:
                    pass



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2016-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import errno
import os
import re
import ssl
import typing
import urllib.parse
import warnings
import hashlib

from . import errors
from . import credentials
from . import platform


EDGEDB_PORT = 5656
ERRNO_RE = re.compile(r"\[Errno (\d+)\]")
TEMPORARY_ERRORS = (
    ConnectionAbortedError,
    ConnectionRefusedError,
    ConnectionResetError,
    FileNotFoundError,
)
TEMPORARY_ERROR_CODES = frozenset({
    errno.ECONNREFUSED,
    errno.ECONNABORTED,
    errno.ECONNRESET,
    errno.ENOENT,
})


class ConnectionParameters(typing.NamedTuple):

    user: str
    password: str
    database: str
    connect_timeout: float
    server_settings: typing.Mapping[str, str]
    ssl_ctx: ssl.SSLContext


class ClientConfiguration(typing.NamedTuple):

    connect_timeout: float
    command_timeout: float
    wait_until_available: float


def _validate_port_spec(hosts, port):
    if isinstance(port, list):
        # If there is a list of ports, its length must
        # match that of the host list.
        if len(port) != len(hosts):
            raise errors.InterfaceError(
                'could not match {} port numbers to {} hosts'.format(
                    len(port), len(hosts)))
    else:
        port = [port for _ in range(len(hosts))]

    return port


def _parse_hostlist(hostlist, port):
    if ',' in hostlist:
        # A comma-separated list of host addresses.
        hostspecs = hostlist.split(',')
    else:
        hostspecs = [hostlist]

    hosts = []
    hostlist_ports = []

    if not port:
        portspec = os.environ.get('EDGEDB_PORT')
        if portspec:
            if ',' in portspec:
                default_port = [int(p) for p in portspec.split(',')]
            else:
                default_port = int(portspec)
        else:
            default_port = EDGEDB_PORT

        default_port = _validate_port_spec(hostspecs, default_port)

    else:
        port = _validate_port_spec(hostspecs, port)

    for i, hostspec in enumerate(hostspecs):
        addr, _, hostspec_port = hostspec.partition(':')
        hosts.append(addr)

        if not port:
            if hostspec_port:
                hostlist_ports.append(int(hostspec_port))
            else:
                hostlist_ports.append(default_port[i])

    if not port:
        port = hostlist_ports

    return hosts, port


def _stash_path(path):
    path = os.path.realpath(path)
    if platform.IS_WINDOWS and not path.startswith('\\\\'):
        path = '\\\\?\\' + path
    hash = hashlib.sha1(str(path).encode('utf-8')).hexdigest()
    base_name = os.path.basename(path)
    dir_name = base_name + '-' + hash
    return platform.search_config_dir('projects', dir_name)


def _parse_verify_hostname(val: str) -> bool:
    val = val.lower()
    if val in {"1", "yes", "true", "y", "t", "on"}:
        return True
    elif val in {"0", "no", "false", "n", "f", "off"}:
        return False
    else:
        raise ValueError(
            "tls_verify_hostname can only be one of yes/no"
        )


def _parse_connect_dsn_and_args(*, dsn, host, port, user,
                                password, database, admin,
                                tls_ca_file, tls_verify_hostname,
                                connect_timeout, server_settings):
    using_credentials = False
    tls_ca_data = None

    if admin:
        warnings.warn(
            'The "admin=True" parameter is deprecated and is scheduled to be '
            'removed. Admin socket should never be used in applications. '
            'Use command-line tool `edgedb` to setup proper credentials.',
            DeprecationWarning, 4)

    if not (
            dsn or host or port or
            os.getenv("EDGEDB_HOST") or os.getenv("EDGEDB_PORT")
    ):
        instance_name = os.getenv("EDGEDB_INSTANCE")
        if instance_name:
            dsn = instance_name
        else:
            dir = os.getcwd()
            if not os.path.exists(os.path.join(dir, 'edgedb.toml')):
                raise errors.ClientConnectionError(
                    f'no `edgedb.toml` found and '
                    f'no connection options specified'
                )
            stash_dir = _stash_path(dir)
            if os.path.exists(stash_dir):
                with open(os.path.join(stash_dir, 'instance-name'), 'rt') as f:
                    dsn = f.read().strip()
            else:
                raise errors.ClientConnectionError(
                    f'Found `edgedb.toml` but the project is not initialized. '
                    f'Run `edgedb project init`.'
                )

    if dsn and dsn.startswith(("edgedb://", "edgedbadmin://")):
        parsed = urllib.parse.urlparse(dsn)

        if parsed.scheme not in ('edgedb', 'edgedbadmin'):
            raise ValueError(
                f'invalid DSN: scheme is expected to be '
                f'"edgedb" or "edgedbadmin", got {parsed.scheme!r}')

        if parsed.scheme == 'edgedbadmin':
            warnings.warn(
                'The `edgedbadmin` scheme is deprecated and is scheduled '
                'to be removed. Admin socket should never be used in '
                'applications. Use command-line tool `edgedb` to setup '
                'proper credentials.',
                DeprecationWarning, 4)

        if admin is None:
            admin = parsed.scheme == 'edgedbadmin'

        if not host and parsed.netloc:
            if '@' in parsed.netloc:
                auth, _, hostspec = parsed.netloc.partition('@')
            else:
                hostspec = parsed.netloc

            if hostspec:
                host, port = _parse_hostlist(hostspec, port)

        if parsed.path and database is None:
            database = parsed.path
            if database.startswith('/'):
                database = database[1:]

        if parsed.username and user is None:
            user = parsed.username

        if parsed.password and password is None:
            password = parsed.password

        if parsed.query:
            query = urllib.parse.parse_qs(parsed.query, strict_parsing=True)
            for key, val in query.items():
                if isinstance(val, list):
                    query[key] = val[-1]

            if 'port' in query:
                val = query.pop('port')
                if not port and val:
                    port = [int(p) for p in val.split(',')]

            if 'host' in query:
                val = query.pop('host')
                if not host and val:
                    host, port = _parse_hostlist(val, port)

            if 'dbname' in query:
                val = query.pop('dbname')
                if database is None:
                    database = val

            if 'database' in query:
                val = query.pop('database')
                if database is None:
                    database = val

            if 'user' in query:
                val = query.pop('user')
                if user is None:
                    user = val

            if 'password' in query:
                val = query.pop('password')
                if password is None:
                    password = val

            if 'tls_ca_file' in query:
                val = query.pop('tls_ca_file')
                if tls_ca_file is None:
                    tls_ca_file = val

            if 'tls_verify_hostname' in query:
                val = query.pop('tls_verify_hostname')
                if tls_verify_hostname is None:
                    tls_verify_hostname = _parse_verify_hostname(val)

            if query:
                if server_settings is None:
                    server_settings = query
                else:
                    server_settings = {**query, **server_settings}
    elif dsn:
        if not dsn.isidentifier():
            raise ValueError(
                f"dsn {dsn!r} is neither a edgedb:// URI "
                f"nor valid instance name"
            )

        using_credentials = True
        path = credentials.get_credentials_path(dsn)
        try:
            creds = credentials.read_credentials(path)
        except Exception as e:
            raise errors.ClientError(
                f"cannot read credentials of instance {dsn!r}"
            ) from e

        if port is None:
            port = creds['port']
        if user is None:
            user = creds['user']
        if host is None and 'host' in creds:
            host = creds['host']
        if password is None and 'password' in creds:
            password = creds['password']
        if database is None and 'database' in creds:
            database = creds['database']
        if tls_ca_file is None and 'tls_cert_data' in creds:
            tls_ca_data = creds['tls_cert_data']
        if tls_verify_hostname is None and 'tls_verify_hostname' in creds:
            tls_verify_hostname = creds['tls_verify_hostname']

    if not host:
        hostspec = os.environ.get('EDGEDB_HOST')
        if hostspec:
            host, port = _parse_hostlist(hostspec, port)

    if not host:
        if platform.IS_WINDOWS or using_credentials:
            host = []
        else:
            host = ['/run/edgedb', '/var/run/edgedb']

        if not admin:
            host.append('localhost')

    if not isinstance(host, list):
        host = [host]

    if not port:
        portspec = os.environ.get('EDGEDB_PORT')
        if portspec:
            if ',' in portspec:
                port = [int(p) for p in portspec.split(',')]
            else:
                port = int(portspec)
        else:
            port = EDGEDB_PORT

    elif isinstance(port, (list, tuple)):
        port = [int(p) for p in port]

    else:
        port = int(port)

    port = _validate_port_spec(host, port)

    if user is None:
        user = os.getenv('EDGEDB_USER')
        if not user:
            user = 'edgedb'

    if password is None:
        password = os.getenv('EDGEDB_PASSWORD')

    if database is None:
        database = os.getenv('EDGEDB_DATABASE')

    if database is None:
        database = 'edgedb'

    if user is None:
        raise errors.InterfaceError(
            'could not determine user name to connect with')

    if database is None:
        raise errors.InterfaceError(
            'could not determine database name to connect to')

    have_unix_sockets = False
    addrs = []
    for h, p in zip(host, port):
        if h.startswith('/'):
            # UNIX socket name
            if '.s.EDGEDB.' not in h:
                if admin:
                    sock_name = f'.s.EDGEDB.admin.{p}'
                else:
                    sock_name = f'.s.EDGEDB.{p}'
                h = os.path.join(h, sock_name)
                have_unix_sockets = True
            addrs.append(h)
        elif not admin:
            # TCP host/port
            addrs.append((h, p))

    if admin and not have_unix_sockets:
        raise ValueError(
            'admin connections are only supported over UNIX sockets')

    if not addrs:
        raise ValueError(
            'could not determine the database address to connect to')

    if server_settings is not None and (
            not isinstance(server_settings, dict) or
            not all(isinstance(k, str) for k in server_settings) or
            not all(isinstance(v, str) for v in server_settings.values())):
        raise ValueError(
            'server_settings is expected to be None or '
            'a Dict[str, str]')

    if admin:
        ssl_ctx = None
    else:
        ssl_ctx = ssl.SSLContext()
        ssl_ctx.verify_mode = ssl.CERT_REQUIRED
        if tls_ca_file or tls_ca_data:
            ssl_ctx.load_verify_locations(
                cafile=tls_ca_file, cadata=tls_ca_data
            )
            if tls_verify_hostname is None:
                tls_verify_hostname = False
        else:
            ssl_ctx.load_default_certs(ssl.Purpose.SERVER_AUTH)
            if platform.IS_WINDOWS:
                import certifi
                ssl_ctx.load_verify_locations(cafile=certifi.where())
            if tls_verify_hostname is None:
                tls_verify_hostname = True
        ssl_ctx.check_hostname = tls_verify_hostname
        ssl_ctx.set_alpn_protocols(['edgedb-binary'])

    params = ConnectionParameters(
        user=user,
        password=password,
        database=database,
        connect_timeout=connect_timeout,
        server_settings=server_settings,
        ssl_ctx=ssl_ctx,
    )

    return addrs, params


def parse_connect_arguments(*, dsn, host, port, user, password,
                            database, admin,
                            tls_ca_file, tls_verify_hostname,
                            timeout, command_timeout, wait_until_available,
                            server_settings):

    if command_timeout is not None:
        try:
            if isinstance(command_timeout, bool):
                raise ValueError
            command_timeout = float(command_timeout)
            if command_timeout <= 0:
                raise ValueError
        except ValueError:
            raise ValueError(
                'invalid command_timeout value: '
                'expected greater than 0 float (got {!r})'.format(
                    command_timeout)) from None

    addrs, params = _parse_connect_dsn_and_args(
        dsn=dsn, host=host, port=port, user=user,
        password=password, admin=admin,
        database=database, connect_timeout=timeout,
        tls_ca_file=tls_ca_file, tls_verify_hostname=tls_verify_hostname,
        server_settings=server_settings,
    )

    config = ClientConfiguration(
        connect_timeout=timeout,
        command_timeout=command_timeout,
        wait_until_available=wait_until_available or 0,
    )

    return addrs, params, config


def check_alpn_protocol(ssl_obj):
    if ssl_obj.selected_alpn_protocol() != 'edgedb-binary':
        raise errors.ClientConnectionFailedError(
            "The server doesn't support the edgedb-binary protocol."
        )


def render_client_no_connection_error(prefix, addr, attempts, duration):
    if isinstance(addr, str):
        msg = (
            f'{prefix}'
            f'\n\tAfter {attempts} attempts in {duration:.1f} sec'
            f'\n\tIs the server running locally and accepting '
            f'\n\tconnections on Unix domain socket {addr!r}?'
        )
    else:
        msg = (
            f'{prefix}'
            f'\n\tAfter {attempts} attempts in {duration:.1f} sec'
            f'\n\tIs the server running on host {addr[0]!r} '
            f'and accepting '
            f'\n\tTCP/IP connections on port {addr[1]}?'
        )
    return msg


def _extract_errno(s):
    """Extract multiple errnos from error string

    When we connect to a host that has multiple underlying IP addresses, say
    ``localhost`` having ``::1`` and ``127.0.0.1``, we get
    ``OSError("Multiple exceptions:...")`` error without ``.errno`` attribute
    set. There are multiple ones in the text, so we extract all of them.
    """
    result = []
    for match in ERRNO_RE.finditer(s):
        result.append(int(match.group(1)))
    if result:
        return result


def wrap_error(e):
    message = str(e)
    if e.errno is None:
        errnos = _extract_errno(message)
    else:
        errnos = [e.errno]

    if errnos:
        is_temp = any((code in TEMPORARY_ERROR_CODES for code in errnos))
    else:
        is_temp = isinstance(e, TEMPORARY_ERRORS)

    if is_temp:
        return errors.ClientConnectionFailedTemporarilyError(message)
    else:
        return errors.ClientConnectionFailedError(message)
